%Bayes, Markov, and Ito:  Linear operators for control of stochastic
%systems
%
%  Terence Sanger
%
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Damped spring model shown in figures 4 and 5.
%%
% model is \dot(p)(v) = Lp(v)
%          \dot(p)(x) = Lp(x)
%
%  dv = [(k/m)(x0-x) -bv] dt + dB1
%  dx = v dt + dB2
%
%       dB1 and dB2 are white noise with variance sigma_x and sigma_v is independent
% this model calculates the joint density p(v,x) assuming independence
%    Alternating estimation of the conditionals densities  p(x|v), then p(v|x)

npoints = 100;  %probability quantization
ntimes = 1000;  %time points.  Total time = ntimes * dt sec.

%boundary values for x and v
min_x = 0;
max_x = 100;
min_v = -100;
max_v = 100;
x = linspace(min_x, max_x, npoints)';
v = linspace(min_v, max_v, npoints)';
dx = (max_x - min_x) / npoints;     %spatial constants for integration
dv = (max_v - min_v) / npoints;

%initial conditions
pvx = zeros(npoints,npoints);       %each row is p(x|v)
mx = (1:npoints) - (npoints/2);
pvx(npoints/2,:) = exp(-0.5*mx.^2/400);  %gaussian initial distribution
pvx = pvx / sum(sum(pvx));          %normalize
x0 = 0.25 * max_x;                  %target

%model parameters
stiffness = 1.5;    
damping = 1;      
sigma_v = 20;                  %noise variance
sigma_x = 20;
dt = 0.01;                   %sec precision for integration
mass = 1;
damping_ratio = damping / (2*sqrt(stiffness * mass));


%proportional control  f = k(x0-x) - bv
k = stiffness * (x0 - x)/mass; %stiffness term converted from force to acceleration
b = damping * v / mass;  %damping term


%save densities for display
Vdensity = zeros(npoints, ntimes);
Xdensity = zeros(size(Vdensity));
clear F;    %movie


figure
for t=1:ntimes,
    
    %first calculate p(v|x=xi) for all xi 
    gv = sigma_v * ones(size(v));           %variance is same for all x (for now)
    for xi = 1:npoints,
        fv = k(xi) * ones(size(v)) - b;     %p(vdot | x=x_i);  damping is same for all x
        dpv = Lp(pvx(:,xi), fv/dv, gv/dv);  %Fokker-Planck operator
        dpv(1) = 0;                         %stabilize the boundaries
        dpv(npoints)=0;
        pvx(:,xi) = pvx(:,xi) + dpv * dt;   %integrate
    end
    
    %now, calculate p(x|v=vi) for all vi
    gx = sigma_x * ones(size(x));
    for vi = 1:npoints,
        fx = v(vi) * ones(size(x));        
        dpx = Lp(pvx(vi,:)', fx/dx, gx/dx);
        dpx(1) = 0;
        dpx(npoints)=0;
        pvx(vi,:) = pvx(vi,:) + dpx' * dt;
    end

    %maintain density properties and compensate for roundoff errors
    pvx(pvx<0) = 0;
    pvx = pvx / sum(sum(pvx));
    
    %save marginals for display
    Xdensity(:,t) = sum(pvx);
    Vdensity(:,t) = sum(pvx');
    
    %show movie of the joint density p(v,x)
    if ~mod(t,10)
        imagesc(pvx*512);
        colormap('gray');
        F(t/10) = getframe;
    end
end

%
%  Figure 4
%
scrsz = get(0,'ScreenSize');
figure('Position',[scrsz(3)/2 scrsz(4)/2 scrsz(3)/2 2*scrsz(3)/10], 'PaperPositionMode', 'auto')
for i=1:10,
    r = ceil(i/5);
    c = i - 5 * (r-1);
    subplot('position', [0.2*(c-1)+.01 0.5*(2-r)+0.03 0.18 0.4]);
    imagesc(frame2im(F(5+10*(i-1))));
    axis square
    axis tight
    axis off
    title([num2str(i) ' sec'], 'fontsize', 12, 'fontweight', 'bold');
end


%
%  Figure 5
%
figure('Position',[scrsz(3)/2 1 scrsz(3)/2 scrsz(3)/4], 'PaperPositionMode', 'auto')
subplot(1,2,1)
imagesc(Xdensity*1024);
colormap('gray'); 
xlabel('time (sec)', 'fontsize', 12, 'fontweight', 'bold');
ylabel('position (m)', 'fontsize', 12, 'fontweight', 'bold');
set(gca, 'XTick', 0:ntimes/2:ntimes);
set(gca, 'XTickLabel', {0, ntimes/200, ntimes/100});
set(gca, 'YTick', [1 npoints/2 npoints]);
set(gca, 'YTickLabel', {'0', '50', '100'});


subplot(1,2,2)
image(Vdensity*1024);
colormap('gray'); 
xlabel('time (sec)', 'fontsize', 12, 'fontweight', 'bold');
ylabel('velocity (m/sec)', 'fontsize', 12, 'fontweight', 'bold');
set(gca, 'XTick', 0:ntimes/2:ntimes);
set(gca, 'XTickLabel', {0, ntimes/200, ntimes/100});
set(gca, 'YTick', [1 npoints/2 npoints]);
set(gca, 'YTickLabel', {'-100', '0', '100'});





